/*
 * braid.cc -- A class for creating and drawing (mathematical) braids in ePiX.
 *
 * Last Change: October 15, 2007
 */

/*
 * Copyright (C) 2007
 * Andrew D. Hwang <ahwang -at- mathcs -dot- holycross -dot- edu>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "epix/constants.h"
#include "epix/spline_data.h"

#include "braid.h"

namespace ePiX {

  static P to_polar(const P& arg)
  {
    return polar(arg.x1(), arg.x2());
  }

  void polar_spline(const P& ctr, const P& pt1, const P& pt2,
		    const P& pt3, const P& pt4, unsigned int num_pts)
  {
    const double dt(1.0/num_pts);
    path spl;
    for (unsigned int i=0; i <= num_pts; ++i)
      spl.pt(ctr + to_polar(spl_pt(pt1, pt2, pt3, pt4, i*dt)));

    spl.draw();
  }

  braid::braid(unsigned int strands)
    : m_strands(strands) { }

  // strands numbered 0, 1, ..., n-1
  braid& braid::twist(unsigned int i1, unsigned int i2, int twists)
  {
    if (0 < twists)
      for (int i=0; i < twists; ++i)
	m_twists.push_back(twist_data(i1 % m_strands, i2 % m_strands));

    else if (twists < 0)
      for (int i=0; i < -twists; ++i)
	m_twists.push_back(twist_data(i2 % m_strands, i1 % m_strands));

    return *this;
  }

  // rectangular
  void braid::draw(const P& pt1, const P& pt2) const
  {
    // check for degenerate rectangle
    if (fabs(pt1.x1() - pt2.x1()) < EPIX_EPSILON
	|| fabs(pt1.x2() - pt2.x2()) < EPIX_EPSILON)
      return;

    // else
    P dx((1.0/m_twists.size())*(pt2 - pt1)&E_1);
    P dy((1.0/m_strands)*(pt2 - pt1)&E_2);
    if ((pt2.x1() - pt1.x1())*(pt2.x2() - pt1.x2()) < 0) // swap roles
      {
	dx = (1.0/m_twists.size())*((pt2 - pt1)&E_2);
	dy = (1.0/m_strands)*((pt2 - pt1)&E_1);
      }

    unsigned int count(0);

    for (std::list<twist_data>::const_iterator p=m_twists.begin();
	 p != m_twists.end(); ++p)
      {
	unsigned int i1((*p).first()), i2((*p).second());
	// undercrossing
	spline(pt1 + count*dx + i2*dy, pt1 + (count+0.5)*dx + i2*dy,
	       pt1 + (count+0.5)*dx + i1*dy, pt1 + (count+1)*dx + i1*dy, 20);

	// straight strands
	for (unsigned int j=0; j < m_strands; ++j)
	  if (j != i1 && j != i2)
	    line(pt1 + count*dx + j*dy, pt1 + (count+1)*dx + j*dy);

	// overcrossing
	spline(pt1 + count*dx + i1*dy, pt1 + (count+0.5)*dx + i1*dy,
	       pt1 + (count+0.5)*dx + i2*dy, pt1 + (count+1)*dx + i2*dy, 20);

	++count;
      }
  }

  // annular
  void braid::draw(const P& ctr, double inner, double outer) const
  {
    const double  in(min(fabs(inner), fabs(outer)));
    const double out(max(fabs(inner), fabs(outer)));

    if (in*out < EPIX_EPSILON || (out - in) < EPIX_EPSILON)
      return;

    // else
    const double dr((out-in)/m_strands);
    const double dth(full_turn()/m_twists.size());
    const unsigned int num_pts((unsigned int) max(20, ceil(120.0/m_twists.size())));

    const P dx(0, dth); // horizontal/angular
    const P dy(dr, 0);  // vertical/radial

    const P pt(in, 0);

    unsigned int count(0);

    for (std::list<twist_data>::const_iterator p=m_twists.begin();
	 p != m_twists.end(); ++p)
      {
	unsigned int i1((*p).first()), i2((*p).second());
	// undercrossing
	polar_spline(ctr, pt + count*dx + i2*dy, pt + (count+0.5)*dx + i2*dy,
		     pt + (count+0.5)*dx + i1*dy, pt + (count+1)*dx + i1*dy,
		     num_pts);

	// straight strands
	for (unsigned int j=0; j < m_strands; ++j)
	  if (j != i1 && j != i2)
	    arc(ctr, in + j*dr, count*dth, (count+1)*dth);

	// overcrossing
	polar_spline(ctr, pt + count*dx + i1*dy, pt + (count+0.5)*dx + i1*dy,
		     pt + (count+0.5)*dx + i2*dy, pt + (count+1)*dx + i2*dy,
		     num_pts);

	++count;
      }
  }


  braid::twist_data::twist_data(unsigned int i1, unsigned int i2)
    : m_strand1(i1), m_strand2(i2) { }

  unsigned int braid::twist_data::first() const
  {
    return m_strand1;
  }

  unsigned int braid::twist_data::second() const
  {
    return m_strand2;
  }
}
